# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
import sys

try:
    from argcomplete.completers import DirectoriesCompleter
except ImportError:
    def DirectoriesCompleter():
        return None
try:
    from argcomplete.completers import FilesCompleter
except ImportError:
    def FilesCompleter(*, allowednames, directories):
        return None

from sros2.api import _artifact_generation
import sros2.errors
from sros2.verb import VerbExtension


class GenerateArtifactsVerb(VerbExtension):
    """Generate keys and permission files from a list of identities and policy files."""

    def add_arguments(self, parser, cli_name) -> None:
        arg = parser.add_argument(
            '-k', '--keystore-root-path', type=pathlib.Path, help='root path of keystore')
        arg.completer = DirectoriesCompleter()
        parser.add_argument(
            '-e', '--enclaves', nargs='*', default=[],
            help='list of identities, aka ROS security enclave names')
        arg = parser.add_argument(
            '-p', '--policy-files', nargs='*', type=pathlib.Path, default=[],
            help='list of policy xml file paths')
        arg.completer = FilesCompleter(
            allowednames=('xml'), directories=False)

    def main(self, *, args) -> int:
        try:
            _artifact_generation.generate_artifacts(
                args.keystore_root_path, args.enclaves, args.policy_files)
        except sros2.errors.SROS2Error as e:
            print(f'Unable to generate artifacts: {str(e)}', file=sys.stderr)
            return 1
        return 0



# Copyright 2019-2020 Canonical Ltd
# Copyright 2016-2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pathlib
from typing import Set

from cryptography import x509
from cryptography.hazmat.backends import default_backend as cryptography_backend
from cryptography.hazmat.primitives import serialization

from rclpy.exceptions import InvalidNamespaceException
from rclpy.validate_namespace import validate_namespace

from sros2 import _utilities
from sros2.api import _policy
import sros2.errors
from sros2.policy import get_policy_default

from . import _keystore, _permission


def create_enclave(keystore_path: pathlib.Path, identity: str) -> None:
    if not _keystore.is_valid_keystore(keystore_path):
        raise sros2.errors.InvalidKeystoreError(keystore_path)

    if not _is_enclave_name_valid(identity):
        raise sros2.errors.InvalidEnclaveNameError(identity)

    relative_path = os.path.normpath(identity.lstrip('/'))
    key_dir = _keystore.get_keystore_enclaves_dir(keystore_path).joinpath(relative_path)
    os.makedirs(key_dir, exist_ok=True)

    # symlink the CA cert in there
    public_certs = ['identity_ca.cert.pem', 'permissions_ca.cert.pem']
    for public_cert in public_certs:
        dst = key_dir.joinpath(public_cert)
        keystore_ca_cert_path = _keystore.get_keystore_public_dir(
            keystore_path).joinpath(public_cert)
        relativepath = pathlib.Path(
            os.path.relpath(keystore_ca_cert_path, key_dir))
        _utilities.create_symlink(src=relativepath, dst=dst)

    # symlink the governance file in there
    keystore_governance_path = _keystore.get_keystore_enclaves_dir(
        keystore_path).joinpath('governance.p7s')
    dest_governance_path = key_dir.joinpath('governance.p7s')
    relativepath = pathlib.Path(
        os.path.relpath(keystore_governance_path, key_dir))
    _utilities.create_symlink(src=relativepath, dst=dest_governance_path)

    keystore_identity_ca_cert_path = _keystore.get_keystore_public_dir(
        keystore_path).joinpath('identity_ca.cert.pem')
    keystore_identity_ca_key_path = _keystore.get_keystore_private_dir(
        keystore_path).joinpath('identity_ca.key.pem')

    # Only create certs/keys if they don't already exist
    cert_path = key_dir.joinpath('cert.pem')
    key_path = key_dir.joinpath('key.pem')
    if not cert_path.is_file() or not key_path.is_file():
        _create_key_and_cert(
            keystore_identity_ca_cert_path,
            keystore_identity_ca_key_path,
            identity,
            cert_path,
            key_path
        )

    # create a wildcard permissions file for this node which can be overridden
    # later using a policy if desired
    policy_file_path = get_policy_default('policy.xml')
    policy_element = _policy.get_policy('/', policy_file_path)
    enclave_element = policy_element.find('enclaves/enclave')
    enclave_element.attrib['path'] = identity

    permissions_path = key_dir.joinpath('permissions.xml')
    _permission.create_permission_file(permissions_path, _utilities.domain_id(), policy_element)

    signed_permissions_path = key_dir.joinpath('permissions.p7s')
    keystore_permissions_ca_cert_path = _keystore.get_keystore_public_dir(
        keystore_path).joinpath('permissions_ca.cert.pem')
    keystore_permissions_ca_key_path = _keystore.get_keystore_private_dir(
        keystore_path).joinpath('permissions_ca.key.pem')
    _utilities.create_smime_signed_file(
        keystore_permissions_ca_cert_path,
        keystore_permissions_ca_key_path,
        permissions_path,
        signed_permissions_path
    )


def get_enclaves(keystore_path: pathlib.Path) -> Set[str]:
    if not _keystore.is_valid_keystore(keystore_path):
        raise sros2.errors.InvalidKeystoreError(keystore_path)

    enclaves_path = _keystore.get_keystore_enclaves_dir(keystore_path)
    enclaves: Set[str] = set()
    if enclaves_path.is_dir():
        key_file_paths = enclaves_path.glob('**/key.pem')

        for key_file_path in key_file_paths:
            enclaves.add(f'/{key_file_path.parent.relative_to(enclaves_path).as_posix()}')

    return enclaves


def _is_enclave_name_valid(name: str) -> bool:
    # TODO(ivanpauno): Use validate_enclave_name when it's propagated to `rclpy`.
    #   This is not to bad for the moment.
    #   Related with https://github.com/ros2/rclpy/issues/528.
    try:
        return validate_namespace(name)
    except InvalidNamespaceException as e:
        print(e)
        return False


def _create_key_and_cert(
        keystore_ca_cert_path: pathlib.Path,
        keystore_ca_key_path: pathlib.Path,
        identity: str,
        cert_path: pathlib.Path,
        key_path: pathlib.Path):
    # Load the CA cert and key from disk
    ca_cert = _utilities.load_cert(keystore_ca_cert_path)

    with open(keystore_ca_key_path, 'rb') as f:
        ca_key = serialization.load_pem_private_key(f.read(), None, cryptography_backend())

    cert, private_key = _utilities.build_key_and_cert(
        x509.Name([x509.NameAttribute(x509.oid.NameOID.COMMON_NAME, identity)]),
        issuer_name=ca_cert.subject,
        ca_key=ca_key)

    _utilities.write_key(private_key, key_path)
    _utilities.write_cert(cert, cert_path)



# Copyright 2016-2017 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
import sys

try:
    from argcomplete.completers import DirectoriesCompleter
except ImportError:
    def DirectoriesCompleter():
        return None
try:
    from argcomplete.completers import FilesCompleter
except ImportError:
    def FilesCompleter(*, allowednames, directories):
        return None

import sros2.keystore
from sros2.verb import VerbExtension


class CreatePermissionVerb(VerbExtension):
    """Create permission."""

    def add_arguments(self, parser, cli_name) -> None:
        arg = parser.add_argument('ROOT', type=pathlib.Path, help='root path of keystore')
        arg.completer = DirectoriesCompleter()
        parser.add_argument('NAME', help='key name, aka ROS enclave name')
        arg = parser.add_argument(
            'POLICY_FILE_PATH', type=pathlib.Path, help='path of the policy xml file')
        arg.completer = FilesCompleter(
            allowednames=('xml'), directories=False)

    def main(self, *, args) -> int:
        try:
            sros2.keystore.create_permission(args.ROOT, args.NAME, args.POLICY_FILE_PATH)
        except sros2.errors.SROS2Error as e:
            print(f'Unable to create permission: {str(e)}', file=sys.stderr)
            return 1
        return 0



# Copyright 2016-2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pathlib

from lxml import etree

from rclpy.utilities import get_rmw_implementation_identifier

from sros2 import _utilities
from sros2.api import _policy
import sros2.errors
from sros2.policy import get_transport_schema, get_transport_template

from . import _keystore


_RMW_WITH_ROS_GRAPH_INFO_TOPIC = (
    'rmw_connextdds',
    'rmw_cyclonedds_cpp',
    'rmw_fastrtps_cpp',
    'rmw_fastrtps_dynamic_cpp'
)


def create_permission(
        keystore_path: pathlib.Path,
        identity: str,
        policy_file_path: pathlib.Path) -> None:
    policy_element = _policy.get_policy(identity, policy_file_path)
    create_permissions_from_policy_element(keystore_path, identity, policy_element)


def create_permissions_from_policy_element(
        keystore_path: pathlib.Path,
        identity: str,
        policy_element: etree.Element) -> None:
    relative_path = os.path.normpath(identity.lstrip('/'))
    key_dir = _keystore.get_keystore_enclaves_dir(keystore_path).joinpath(relative_path)
    permissions_path = key_dir.joinpath('permissions.xml')
    create_permission_file(permissions_path, _utilities.domain_id(), policy_element)

    signed_permissions_path = os.path.join(key_dir, 'permissions.p7s')
    keystore_permissions_ca_cert_path = os.path.join(
        _keystore.get_keystore_public_dir(keystore_path), 'permissions_ca.cert.pem')
    keystore_permissions_ca_key_path = os.path.join(
        _keystore.get_keystore_private_dir(keystore_path), 'permissions_ca.key.pem')
    _utilities.create_smime_signed_file(
        keystore_permissions_ca_cert_path,
        keystore_permissions_ca_key_path,
        permissions_path,
        signed_permissions_path
    )


def create_permission_file(path: pathlib.Path, domain_id, policy_element) -> None:
    permissions_xsl_path = get_transport_template('dds', 'permissions.xsl')
    permissions_xsl = etree.XSLT(etree.parse(str(permissions_xsl_path)))
    permissions_xsd_path = get_transport_schema('dds', 'permissions.xsd')
    permissions_xsd = etree.XMLSchema(etree.parse(str(permissions_xsd_path)))

    kwargs = {}

    cert_path = path.parent.joinpath('cert.pem')
    cert_content = _utilities.load_cert(cert_path)
    kwargs['not_valid_before'] = etree.XSLT.strparam(cert_content.not_valid_before.isoformat())
    kwargs['not_valid_after'] = etree.XSLT.strparam(cert_content.not_valid_after.isoformat())

    if get_rmw_implementation_identifier() in _RMW_WITH_ROS_GRAPH_INFO_TOPIC:
        kwargs['allow_ros_discovery_topic'] = etree.XSLT.strparam('1')
    permissions_xml = permissions_xsl(policy_element, **kwargs)

    domain_id_elements = permissions_xml.findall('permissions/grant/*/domains/id')
    for domain_id_element in domain_id_elements:
        domain_id_element.text = domain_id

    try:
        permissions_xsd.assertValid(permissions_xml)
    except etree.DocumentInvalid as e:
        raise sros2.errors.InvalidPermissionsXMLError(e) from e

    with open(path, 'wb') as f:
        f.write(etree.tostring(permissions_xml, pretty_print=True))



# Copyright 2020 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Workaround to make importlib.resources path lookup work.

importlib.resources in Python 3.8 and earlier doesn't support looking up paths
in subdirectories.  However, it does support looking up paths in python modules,
so we add this __init__.py to make this subdirectoy a module.  This workaround
can be removed when all of the target platforms support Python 3.9 and later;
see https://gitlab.com/python-devs/importlib_resources/issues/58 for details.
"""



# Copyright 2016-2017 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from argcomplete.completers import DirectoriesCompleter
except ImportError:
    def DirectoriesCompleter():
        return None

import pathlib
import sys
import warnings

import sros2.keystore
from sros2.verb import VerbExtension


class ListEnclavesVerb(VerbExtension):
    """List enclaves in keystore."""

    def add_arguments(self, parser, cli_name) -> None:
        arg = parser.add_argument('ROOT', type=pathlib.Path, help='root path of keystore')
        arg.completer = DirectoriesCompleter()

    def main(self, *, args) -> int:
        try:
            enclaves = sros2.keystore.get_enclaves(args.ROOT)
            if enclaves:
                # Print each enclave on its own line
                for enclave in sorted(enclaves):
                    print(enclave)
            return 0
        except sros2.errors.SROS2Error as e:
            print(f'Unable to list enclaves: {str(e)}', file=sys.stderr)
        return 1


class ListKeysVerb(ListEnclavesVerb):
    """DEPRECATED: List enclaves in keystore. Use list_enclaves instead."""

    def main(self, *, args) -> int:
        warnings.warn(
            'list_keys is deprecated and will be removed in a future release. Use list_enclaves '
            'instead.', FutureWarning)
        return super().main(args=args)



# Copyright 2020 Canonical Ltd
# Copyright 2016-2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import os
import pathlib

from cryptography import x509
from cryptography.hazmat.backends import default_backend as cryptography_backend
from cryptography.hazmat.bindings.openssl.binding import Binding as SSLBinding
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec

import sros2.errors

_DOMAIN_ID_ENV = 'ROS_DOMAIN_ID'
_KEYSTORE_DIR_ENV = 'ROS_SECURITY_KEYSTORE'


def create_symlink(*, src: pathlib.Path, dst: pathlib.Path):
    if dst.exists():
        # Don't do more work than we need to
        if dst.samefile(dst.parent.joinpath(src)):
            return
        os.remove(dst)
    os.symlink(src, dst)


def domain_id() -> str:
    return os.getenv(_DOMAIN_ID_ENV, '0')


def get_keystore_path_from_env() -> pathlib.Path:
    root_keystore_path = os.getenv(_KEYSTORE_DIR_ENV)
    if root_keystore_path is None:
        raise sros2.errors.InvalidKeystoreEnvironmentError(_KEYSTORE_DIR_ENV)
    return pathlib.Path(root_keystore_path)


def create_smime_signed_file(cert_path, key_path, unsigned_file_path, signed_file_path):
    # Load the CA cert and key from disk
    cert = load_cert(cert_path)

    with open(key_path, 'rb') as key_file:
        private_key = serialization.load_pem_private_key(
            key_file.read(), None, cryptography_backend())

    # Get the contents of the unsigned file, which we're about to sign
    with open(unsigned_file_path, 'rb') as f:
        content = f.read()

    # Sign the contents, and write the result to the appropriate place
    with open(signed_file_path, 'wb') as f:
        f.write(_sign_bytes(cert, private_key, content))


def build_key_and_cert(subject_name, *, ca=False, ca_key=None, issuer_name=''):
    if not issuer_name:
        issuer_name = subject_name

    # DDS-Security section 9.3.1 calls for prime256v1, for which SECP256R1 is an alias
    private_key = ec.generate_private_key(ec.SECP256R1, cryptography_backend())
    if not ca_key:
        ca_key = private_key

    if ca:
        extension = x509.BasicConstraints(ca=True, path_length=1)
    else:
        extension = x509.BasicConstraints(ca=False, path_length=None)

    utcnow = datetime.datetime.utcnow()
    builder = x509.CertificateBuilder(
        ).issuer_name(
            issuer_name
        ).serial_number(
            x509.random_serial_number()
        ).not_valid_before(
            # Using a day earlier here to prevent Connext (5.3.1) from complaining
            # when extracting it from the permissions file and thinking it's in the future
            # https://github.com/ros2/ci/pull/436#issuecomment-624874296
            utcnow - datetime.timedelta(days=1)
        ).not_valid_after(
            # TODO: This should not be hard-coded
            utcnow + datetime.timedelta(days=3650)
        ).public_key(
            private_key.public_key()
        ).subject_name(
            subject_name
        ).add_extension(
            extension, critical=ca
        )
    cert = builder.sign(ca_key, hashes.SHA256(), cryptography_backend())

    return (cert, private_key)


def write_key(
    key,
    key_path: pathlib.Path,
    *,
    encoding=serialization.Encoding.PEM,
    serialization_format=serialization.PrivateFormat.PKCS8,
    encryption_algorithm=serialization.NoEncryption()
):
    with open(key_path, 'wb') as f:
        f.write(key.private_bytes(
            encoding=encoding,
            format=serialization_format,
            encryption_algorithm=encryption_algorithm))


def write_cert(cert, cert_path: pathlib.Path, *, encoding=serialization.Encoding.PEM):
    with open(cert_path, 'wb') as f:
        f.write(cert.public_bytes(encoding=encoding))


def load_cert(cert_path: pathlib.Path):
    with open(cert_path, 'rb') as cert_file:
        return x509.load_pem_x509_certificate(
            cert_file.read(), cryptography_backend())


def _sign_bytes(cert, key, byte_string):
    # Using two flags here to get the output required:
    #   - PKCS7_DETACHED: Use cleartext signing
    #   - PKCS7_TEXT: Set the MIME headers for text/plain
    flags = SSLBinding.lib.PKCS7_DETACHED
    flags |= SSLBinding.lib.PKCS7_TEXT

    # Convert the byte string into a buffer for SSL
    bio_in = SSLBinding.lib.BIO_new_mem_buf(byte_string, len(byte_string))
    try:
        pkcs7 = SSLBinding.lib.PKCS7_sign(
            cert._x509, key._evp_pkey, SSLBinding.ffi.NULL, bio_in, flags)
    finally:
        # Free the memory allocated for the buffer
        SSLBinding.lib.BIO_free(bio_in)

    # PKCS7_sign consumes the buffer; allocate a new one again to get it into the final document
    bio_in = SSLBinding.lib.BIO_new_mem_buf(byte_string, len(byte_string))
    try:
        # Allocate a buffer for the output document
        bio_out = SSLBinding.lib.BIO_new(SSLBinding.lib.BIO_s_mem())
        try:
            # Write the final document out to the buffer
            SSLBinding.lib.SMIME_write_PKCS7(bio_out, pkcs7, bio_in, flags)

            # Copy the output document back to python-managed memory
            result_buffer = SSLBinding.ffi.new('char**')
            buffer_length = SSLBinding.lib.BIO_get_mem_data(bio_out, result_buffer)
            output = SSLBinding.ffi.buffer(result_buffer[0], buffer_length)[:]
        finally:
            # Free the memory required for the output buffer
            SSLBinding.lib.BIO_free(bio_out)
    finally:
        # Free the memory allocated for the input buffer
        SSLBinding.lib.BIO_free(bio_in)

    return output



# Copyright 2016-2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
from typing import List

from sros2 import _utilities, keystore
from sros2.policy import load_policy

from . import _policy


def generate_artifacts(
        keystore_path: pathlib.Path = None,
        identity_names: List[str] = [],
        policy_files: List[pathlib.Path] = []) -> None:
    if keystore_path is None:
        keystore_path = _utilities.get_keystore_path_from_env()
        if keystore_path is None:
            return False
    if not keystore.is_valid_keystore(keystore_path):
        print('%s is not a valid keystore, creating new keystore' % keystore_path)
        keystore.create_keystore(keystore_path)

    # Create enclaves for all provided identities
    for identity in identity_names:
        keystore.create_enclave(keystore_path, identity)
    for policy_file in policy_files:
        policy_tree = load_policy(policy_file)
        enclaves_element = policy_tree.find('enclaves')
        for enclave in enclaves_element:
            identity_name = enclave.get('path')
            if identity_name not in identity_names:
                keystore.create_enclave(keystore_path, identity_name)
            policy_element = _policy.get_policy_from_tree(identity_name, policy_tree)
            keystore._permission.create_permissions_from_policy_element(
                keystore_path, identity_name, policy_element)



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import urllib.parse
import urllib.request

import ament_index_python


_xml_cache_path = urllib.parse.urljoin(
    'file:',
    urllib.request.pathname2url(
        os.path.join(
            ament_index_python.get_package_share_directory('sros2'),
            'xml_cache',
            'xhtml-cache.xml'
        )
    )
)


if 'XML_CATALOG_FILES' not in os.environ:
    os.environ['XML_CATALOG_FILES'] = _xml_cache_path
elif _xml_cache_path not in os.environ['XML_CATALOG_FILES']:
    os.environ['XML_CATALOG_FILES'] += ' ' + _xml_cache_path



# Copyright 2018-2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib

from lxml import etree

try:
    import importlib.resources as importlib_resources
except ModuleNotFoundError:
    # (clalancette): This fallback is necessary when importlib-resources is installed from pip.
    # Note that we have to ignore the type, otherwise mypy throws a warning.
    # See https://github.com/python/mypy/issues/1153 for details.
    import importlib_resources  # type: ignore

POLICY_VERSION = '0.2.0'


def get_policy_default(name: str) -> pathlib.Path:
    with importlib_resources.path('sros2.policy.defaults', name) as path:
        return path


def get_policy_schema(name: str) -> pathlib.Path:
    with importlib_resources.path('sros2.policy.schemas', name) as path:
        return path


def get_policy_template(name: str) -> pathlib.Path:
    with importlib_resources.path('sros2.policy.templates', name) as path:
        return path


def get_transport_default(transport: str, name: str) -> pathlib.Path:
    module = 'sros2.policy.defaults.' + transport
    with importlib_resources.path(module, name) as path:
        return path


def get_transport_schema(transport: str, name: str) -> pathlib.Path:
    module = 'sros2.policy.schemas.' + transport
    with importlib_resources.path(module, name) as path:
        return path


def get_transport_template(transport: str, name: str) -> pathlib.Path:
    module = 'sros2.policy.templates.' + transport
    with importlib_resources.path(module, name) as path:
        return path


def load_policy(policy_file_path: pathlib.Path) -> etree.ElementTree:
    if not policy_file_path.is_file():
        raise FileNotFoundError("policy file '%s' does not exist" % policy_file_path)
    policy = etree.parse(str(policy_file_path))
    policy.xinclude()
    try:
        policy_xsd_path = get_policy_schema('policy.xsd')
        policy_xsd = etree.XMLSchema(etree.parse(str(policy_xsd_path)))
        policy_xsd.assertValid(policy)
    except etree.DocumentInvalid as e:
        raise RuntimeError(str(e))
    return policy


def dump_policy(policy, stream) -> None:
    policy_xsl_path = get_policy_template('policy.xsl')
    policy_xsl = etree.XSLT(etree.parse(str(policy_xsl_path)))
    policy = policy_xsl(policy)
    try:
        policy_xsd_path = get_policy_schema('policy.xsd')
        policy_xsd = etree.XMLSchema(etree.parse(str(policy_xsd_path)))
        policy_xsd.assertValid(policy)
    except etree.DocumentInvalid as e:
        raise RuntimeError(str(e))
    stream.write(etree.tostring(policy, pretty_print=True).decode())



# Copyright 2016-2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lxml import etree

from sros2.policy import load_policy


def get_policy(name, policy_file_path):
    policy_tree = load_policy(policy_file_path)
    return get_policy_from_tree(name, policy_tree)


def get_policy_from_tree(name, policy_tree):
    enclave_element = policy_tree.find(
        path=f'enclaves/enclave[@path="{name}"]')
    if enclave_element is None:
        raise RuntimeError(f'unable to find enclave "{name}"')
    enclaves_element = etree.Element('enclaves')
    enclaves_element.append(enclave_element)
    policy_element = etree.Element('policy')
    policy_element.append(enclaves_element)
    return policy_element



# Copyright 2020 Canonical Ltd
# Copyright 2016-2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib

from cryptography import x509

import lxml

from sros2 import _utilities
import sros2.errors
from sros2.policy import get_transport_default, get_transport_schema


_KS_ENCLAVES = 'enclaves'
_KS_PUBLIC = 'public'
_KS_PRIVATE = 'private'
_DEFAULT_COMMON_NAME = 'sros2CA'


def create_keystore(keystore_path: pathlib.Path) -> None:
    if is_valid_keystore(keystore_path):
        raise sros2.errors.KeystoreExistsError(keystore_path)

    for path in (
            keystore_path,
            keystore_path.joinpath(_KS_PUBLIC),
            keystore_path.joinpath(_KS_PRIVATE),
            keystore_path.joinpath(_KS_ENCLAVES)):
        path.mkdir(parents=True, exist_ok=True)

    keystore_ca_cert_path = keystore_path.joinpath(_KS_PUBLIC, 'ca.cert.pem')
    keystore_ca_key_path = keystore_path.joinpath(_KS_PRIVATE, 'ca.key.pem')

    keystore_permissions_ca_cert_path = keystore_path.joinpath(
        _KS_PUBLIC, 'permissions_ca.cert.pem')
    keystore_permissions_ca_key_path = keystore_path.joinpath(
        _KS_PRIVATE, 'permissions_ca.key.pem')

    keystore_identity_ca_cert_path = keystore_path.joinpath(
        _KS_PUBLIC, 'identity_ca.cert.pem')
    keystore_identity_ca_key_path = keystore_path.joinpath(
        _KS_PRIVATE, 'identity_ca.key.pem')

    required_files = (
        keystore_permissions_ca_cert_path,
        keystore_permissions_ca_key_path,
        keystore_identity_ca_cert_path,
        keystore_identity_ca_key_path,
    )

    # Create new CA if one doesn't already exist
    if not all(x.is_file() for x in required_files):
        _create_ca_key_cert(keystore_ca_key_path, keystore_ca_cert_path)

        for path in (keystore_permissions_ca_cert_path, keystore_identity_ca_cert_path):
            _utilities.create_symlink(src=pathlib.Path('ca.cert.pem'), dst=path)

        for path in (keystore_permissions_ca_key_path, keystore_identity_ca_key_path):
            _utilities.create_symlink(src=pathlib.Path('ca.key.pem'), dst=path)

    # Create governance file if it doesn't already exist
    gov_path = keystore_path.joinpath(_KS_ENCLAVES, 'governance.xml')
    if not gov_path.is_file():
        _create_governance_file(gov_path, _utilities.domain_id())

    # Sign governance file if it hasn't already been signed
    signed_gov_path = keystore_path.joinpath(_KS_ENCLAVES, 'governance.p7s')
    if not signed_gov_path.is_file():
        _utilities.create_smime_signed_file(
            keystore_permissions_ca_cert_path,
            keystore_permissions_ca_key_path,
            gov_path,
            signed_gov_path)


def is_valid_keystore(path: pathlib.Path) -> bool:
    return (
        path.joinpath(_KS_PUBLIC, 'permissions_ca.cert.pem').is_file() and
        path.joinpath(_KS_PUBLIC, 'identity_ca.cert.pem').is_file() and
        path.joinpath(_KS_PRIVATE, 'permissions_ca.key.pem').is_file() and
        path.joinpath(_KS_PRIVATE, 'identity_ca.key.pem').is_file() and
        path.joinpath(_KS_ENCLAVES, 'governance.p7s').is_file()
    )


def get_keystore_enclaves_dir(keystore_path: pathlib.Path) -> pathlib.Path:
    return keystore_path.joinpath(_KS_ENCLAVES)


def get_keystore_public_dir(keystore_path: pathlib.Path) -> pathlib.Path:
    return keystore_path.joinpath(_KS_PUBLIC)


def get_keystore_private_dir(keystore_path: pathlib.Path) -> pathlib.Path:
    return keystore_path.joinpath(_KS_PRIVATE)


def _create_ca_key_cert(ca_key_out_path, ca_cert_out_path):
    cert, private_key = _utilities.build_key_and_cert(
        x509.Name([x509.NameAttribute(x509.oid.NameOID.COMMON_NAME, _DEFAULT_COMMON_NAME)]),
        ca=True)

    _utilities.write_key(private_key, ca_key_out_path)
    _utilities.write_cert(cert, ca_cert_out_path)


def _create_governance_file(path: pathlib.Path, domain_id: str):
    # for this application we are only looking to authenticate and encrypt;
    # we do not need/want access control at this point.
    governance_xml_path = get_transport_default('dds', 'governance.xml')
    governance_xml = lxml.etree.parse(str(governance_xml_path))

    governance_xsd_path = get_transport_schema('dds', 'governance.xsd')
    governance_xsd = lxml.etree.XMLSchema(lxml.etree.parse(str(governance_xsd_path)))

    domain_id_elements = governance_xml.findall(
        'domain_access_rules/domain_rule/domains/id')
    for domain_id_element in domain_id_elements:
        domain_id_element.text = domain_id

    try:
        governance_xsd.assertValid(governance_xml)
    except lxml.etree.DocumentInvalid as e:
        raise sros2.errors.InvalidGovernanceXMLError(e) from e

    with open(path, 'wb') as f:
        f.write(lxml.etree.tostring(governance_xml, pretty_print=True))



# Copyright 2016-2017 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple
import pathlib
import sys

try:
    from argcomplete.completers import DirectoriesCompleter
except ImportError:
    def DirectoriesCompleter():
        return None
try:
    from argcomplete.completers import FilesCompleter
except ImportError:
    def FilesCompleter(*, allowednames, directories):
        return None

from lxml import etree

from ros2cli.node.strategy import add_arguments as add_strategy_node_arguments
from ros2cli.node.strategy import NodeStrategy

from sros2.policy import (
    dump_policy,
    load_policy,
    POLICY_VERSION,
)

from sros2.verb import VerbExtension


_HIDDEN_NODE_PREFIX = '_'

_NodeName = namedtuple('_NodeName', ('node', 'ns', 'fqn', 'path'))
_TopicInfo = namedtuple('_TopicInfo', ('fqn', 'type'))


class GeneratePolicyVerb(VerbExtension):
    """Generate XML policy file from ROS graph data."""

    def add_arguments(self, parser, cli_name) -> None:
        arg = parser.add_argument(
            'POLICY_FILE_PATH', type=pathlib.Path, help='path of the policy xml file')
        arg.completer = FilesCompleter(
            allowednames=('xml'), directories=False)
        add_strategy_node_arguments(parser)

    def get_policy(self, policy_file_path: pathlib.Path):
        if policy_file_path.is_file():
            return load_policy(policy_file_path)
        else:
            enclaves = etree.Element('enclaves')
            policy = etree.Element('policy')
            policy.attrib['version'] = POLICY_VERSION
            policy.append(enclaves)
            return policy

    def get_profile(self, policy, node_name):
        enclave = policy.find(
            path=f'enclaves/enclave[@path="{node_name.path}"]')
        if enclave is None:
            enclave = etree.Element('enclave')
            enclave.attrib['path'] = node_name.path
            profiles = etree.Element('profiles')
            enclave.append(profiles)
            enclaves = policy.find('enclaves')
            enclaves.append(enclave)
        profile = enclave.find(
            path=f'profiles/profile[@ns="{node_name.ns}"][@node="{node_name.node}"]')
        if profile is None:
            profile = etree.Element('profile')
            profile.attrib['ns'] = node_name.ns
            profile.attrib['node'] = node_name.node
            profiles = enclave.find('profiles')
            profiles.append(profile)
        return profile

    def get_permissions(self, profile, permission_type, rule_type, rule_qualifier):
        permissions = profile.find(
            path=f'{permission_type}s[@{rule_type}="{rule_qualifier}"]')
        if permissions is None:
            permissions = etree.Element(permission_type + 's')
            permissions.attrib[rule_type] = rule_qualifier
            profile.append(permissions)
        return permissions

    def add_permission(
            self, profile, permission_type, rule_type, rule_qualifier, expressions, node_name):
        permissions = self.get_permissions(profile, permission_type, rule_type, rule_qualifier)
        for expression in expressions:
            permission = etree.Element(permission_type)
            if expression.fqn.startswith(node_name.fqn + '/'):
                permission.text = '~' + expression.fqn[len(node_name.fqn):]
            elif expression.fqn.startswith(node_name.ns + '/'):
                permission.text = expression.fqn[len(node_name.ns + '/'):]
            elif expression.fqn.count('/') == 1 and node_name.ns == '/':
                permission.text = expression.fqn[len('/'):]
            else:
                permission.text = expression.fqn
            permissions.append(permission)

    def main(self, *, args) -> int:
        policy = self.get_policy(args.POLICY_FILE_PATH)
        with NodeStrategy(args) as node:
            node_names = _get_node_names(node=node, include_hidden_nodes=False)

            if not len(node_names):
                print('No nodes detected in the ROS graph. No policy file was generated.',
                      file=sys.stderr)
                return 1

            for node_name in node_names:
                profile = self.get_profile(policy, node_name)
                subscribe_topics = _get_subscriber_info(node=node, node_name=node_name)
                if subscribe_topics:
                    self.add_permission(
                        profile, 'topic', 'subscribe', 'ALLOW', subscribe_topics, node_name)
                publish_topics = _get_publisher_info(node=node, node_name=node_name)
                if publish_topics:
                    self.add_permission(
                        profile, 'topic', 'publish', 'ALLOW', publish_topics, node_name)
                reply_services = _get_service_info(node=node, node_name=node_name)
                if reply_services:
                    self.add_permission(
                        profile, 'service', 'reply', 'ALLOW', reply_services, node_name)
                request_services = _get_client_info(node=node, node_name=node_name)
                if request_services:
                    self.add_permission(
                        profile, 'service', 'request', 'ALLOW', request_services, node_name)

        with open(args.POLICY_FILE_PATH, 'w') as stream:
            dump_policy(policy, stream)
        return 0


def _get_node_names(*, node, include_hidden_nodes=False):
    node_names_and_namespaces_with_enclaves = node.get_node_names_and_namespaces_with_enclaves()
    return [
        _NodeName(
            node=t[0],
            ns=t[1],
            fqn=t[1] + ('' if t[1].endswith('/') else '/') + t[0],
            path=t[2])
        for t in node_names_and_namespaces_with_enclaves
        if (
            include_hidden_nodes or
            (t[0] and not t[0].startswith(_HIDDEN_NODE_PREFIX))
        )
    ]


def _get_topics(node_name, func):
    names_and_types = func(node_name.node, node_name.ns)
    return [
        _TopicInfo(
            fqn=t[0],
            type=t[1])
        for t in names_and_types]


def _get_subscriber_info(node, node_name):
    return _get_topics(node_name, node.get_subscriber_names_and_types_by_node)


def _get_publisher_info(node, node_name):
    return _get_topics(node_name, node.get_publisher_names_and_types_by_node)


def _get_service_info(node, node_name):
    return _get_topics(node_name, node.get_service_names_and_types_by_node)


def _get_client_info(node, node_name):
    return _get_topics(node_name, node.get_client_names_and_types_by_node)



# Copyright 2020 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib

import lxml.etree


class SROS2Error(Exception):
    """Base class for all SROS2 exceptions."""


class KeystoreError(SROS2Error):
    """Base class for keystore-related exceptions."""


class InvalidKeystoreError(KeystoreError):
    """Exception raised when the keystore is invalid."""

    def __init__(self, path: pathlib.Path):
        self.path = path
        super().__init__(f"'{str(path)}' is not a valid keystore")


class KeystoreExistsError(KeystoreError):
    """Exception raised when trying to create a keystore that already exists."""

    def __init__(self, path: pathlib.Path):
        self.path = path
        super().__init__(f"keystore '{str(path)}' already exists")


class EnclaveError(SROS2Error):
    """Base class for enclave-related exceptions."""


class InvalidEnclaveNameError(EnclaveError):
    """Exception raised when an enclave name is invalid."""

    def __init__(self, name: str):
        self.name = name
        super().__init__(f"'{name}' is not a valid enclave name")


class XmlError(SROS2Error):
    """Base class for XML-related exceptions."""


class InvalidGovernanceXMLError(XmlError):
    """Exception raised when governance XML is invalid."""

    def __init__(self, error: lxml.etree.DocumentInvalid):
        self.error = error
        super().__init__(f'invalid governance XML: {str(error)}')


class InvalidPermissionsXMLError(XmlError):
    """Exception raised when permissions XML is invalid."""

    def __init__(self, error: lxml.etree.DocumentInvalid):
        self.error = error
        super().__init__(f'invalid permissions XML: {str(error)}')


class SROS2EnvironmentError(SROS2Error):
    """Base class for environment-related exceptions."""


class InvalidKeystoreEnvironmentError(KeystoreError):
    """Exception raised when the keystore environment variable isn't valid."""

    def __init__(self, variable_name: str):
        self.variable_name = variable_name
        super().__init__(f"'{variable_name}' isn't pointing at a valid keystore")
